#!/usr/bin/python
# coding: utf-8

import os, sys                  # for daemonize
import SocketServer, urllib, socket     # for service
import re
import  gettext
import locale
import logging, logging.config # for logging
import codecs                  # for console in/output encoding

options = None
encoding = ''

# utility functions
def isContentOnly(seq, aset):
    for c in seq:
        if c not in aset:
            return False
        pass
    return True

def isContentAny(seq, aset):
    for c in seq:
        if c in aset:
            return True
        pass
    return False

def buildTranslater(*args, **kwds):
    import logging, logging.handlers

    tmp = dict(*args, **kwds)
    rtrns = re.compile('|'.join(map(re.escape, tmp)))
    def translateOne(match):
        return tmp[match.group(0)]
    def translate(text):
        return rtrns.sub(translateOne, text)
    return translate

def setupLogger():
    '''
    loggingモジュールの初期化
    それぞれの出力先は設定ファイルとして与える
    この関数を通過以降は logging を通じてログ出力が可能となる
    '''
    if options.filelog:
        if sys.platform == 'win32':
            logging.config.fileConfig('filelogwin.ini')
            pass
        else:
            logging.config.fileConfig('filelog.ini')
            pass
        pass
    if sys.platform == 'win32':
        if options.ntevent:
            # このログ出力先はログソースの設定などのためHKLMへのレジストリ書き込むため管理者権限が必要
            logging.config.fileConfig('ntevent.ini')
            pass
        pass
    else:
        if options.syslog:
            logging.config.fileConfig('syslog.ini')
            pass
        pass
    if not options.debug:
        logging.getLogger().setLevel(logging.INFO)
        pass
    else:
        logging.getLogger().setLevel(logging.DEBUG)
        pass
    return

def debug_method(arg):
    try:
        logging.debug(u'"%s"\n' % arg.decode(SocialSKKRequestHandler.CLIENT_CHARSET))
    except UnicodeDecodeError:
        try:
            logging.debug(u'"%s"\n' % arg.decode(SocialSKKRequestHandler.SERVER_CHARSET))
        except UnicodeDecodeError:
            logging.debug('"%r"\n' % arg)
            pass
        pass
    except UnicodeEncodeError:
        logging.debug(u'"%s"\n' % arg)
        pass
    return

def parseoptions():
    '''
    起動オプションを解析する
    大域変数 options に解析結果を格納する
    '''
    import optparse
    global options
    parser = optparse.OptionParser()
    parser.add_option('-p', '--port', type='int', dest='port', default=5511, help=_('server port number'))
    parser.add_option('-d', '--debug', action='store_true', dest='debug', default=False, help=_('debugging log to error log file'))
    parser.add_option('-f', '--filelog', action='store_true', dest='filelog', default=False, help=_('log to file'))
    if sys.platform == 'win32':
        parser.add_option('-e', '--eventlog', action='store_true', dest='ntevent', default=False, help=_('log to Event Log'))
        pass
    else:
        # loggingモジュールを使用するようになったのでデーモンの標準入出力は/dev/nullへ送り込む
        #parser.add_option('-l', '--log', type='string', dest='log', default='/tmp/skkserv.log', help=_('log filename'))
        #parser.add_option('-e', '--errorlog', type='string', dest='err', default='/tmp/skkserv.err', help=_('error log filename'))
        parser.add_option('-r', '--pidfile', type='string', dest='run', default='/tmp/skkserv.run', help=_('process id file'))
        parser.add_option('-s', '--syslog', action='store_true', dest='syslog', default=False, help=_('log to syslog'))
        pass
    # 解析対象外となった引数を他の解析処理に行なわせるため
    (options, args) = parser.parse_args()
    sys.argv[1:] = args
    if sys.platform == 'win32':
        if options.filelog and options.ntevent:
            parser.error(_('options -f and -e are mutually exclusive'))
            pass
        elif not (options.filelog or options.ntevent):
            # ログ記録オプションが設定されていない
            # pywin32を使えばサービス化できそうなので、可能になったらEventLogに記録する
            # それまではrootLoggerの規定値を使う
            #options.ntevent = True
            pass
        pass
    if sys.platform != 'win32':
        if options.filelog and options.syslog:
            parser.error(_('options -f and -s are mutually exclusive'))
            pass
        elif not (options.filelog or options.syslog):
            # ログ記録オプションが設定されていない
            # デーモン化しているため標準入出力がどこかへ行ってしまうので、syslogに記録させる
            options.syslog = True
            pass
        pass
    return

class SocialSKKServer(SocketServer.ThreadingTCPServer):
    allow_reuse_address = True
    daemon_threads = True
    pass

class SocialSKKRequestHandler(SocketServer.StreamRequestHandler):
    VERSION = u'PySocialSKKServ0.4.1 '
    SERVER = u'http://www.social-ime.com:80/'
    CLIENT_END = u'0'
    CLIENT_REQUEST = u'1'
    CLIENT_VERSION = u'2'
    CLIENT_HOST = u'3'
    CLIENT_SERVER_COMPLETION = u'4'

    SERVER_ERROR = u'0'
    SERVER_FOUND = u'1'
    SERVER_NOT_FOUND = u'4'
    SERVER_FULL = u'9'

    COMBUFSIZE = 1024

    SERVER_CHARSET = 'EUC-JP'
    CLIENT_CHARSET = 'EUC-JP'

    specialbefore = u'/;#'
    specialafter = {
        u'/': u'\\057',
        u';': u'\\073',
        u'#': u'\\043',
        }
    specialxlat = None

    # str.isalnumだと「ひらがな」なども「アルファベット」扱いなので自作
    def isalnum(self, str):
        return isContentOnly(str, u'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')

    def isHiragana(self, str):
        return isContentOnly(str, u'あいうえおぁぃぅぇぉかきくけこがぎぐげごさしすせそざじずぜぞたちつてとだぢづでどっなにぬねのはひふへほぱぴぷぺぽばびぶべぼまみむめもやゆよゃゅょらりるれろわゐゑをん')

    def isKatakana(self, str):
        return isContentOnly(str, u'アイウエオァィゥェォカキクケコガギグゲゴサシスセソザジズゼゾタチツテトダヂヅデドッナニヌネノハヒフヘホパピプペポバビブベボマミムメモヤユヨャュョラリルレロワヰヱヲン')

    def isKana(self, str):
        return isContentOnly(str, u'あいうえおぁぃぅぇぉかきくけこがぎぐげごさしすせそざじずぜぞたちつてとだぢづでどっなにぬねのはひふへほぱぴぷぺぽばびぶべぼまみむめもやゆよゃゅょらりるれろわゐゑをんアイウエオァィゥェォカキクケコガギグゲゴサシスセソザジズゼゾタチツテトダヂヅデドッナニヌネノハヒフヘホパピプペポバビブベボマミムメモヤユヨャュョラリルレロワヰヱヲン')

    # 変換の統計
    nRequest = 0
    nError = 0
    nOK = 0
    nFail = 0
    nPredict = 0
    nDict = 0

    okurimap = {
        u'a': u'あ',
        u'i': u'い',
        u'u': u'う',
        u'e': u'え',
        u'o': u'お',
        u'k': u'かきくけこ',
        u'g': u'がぎぐげご',
        u's': u'さしすせそ',
        u'z': u'ざじずぜぞ',
        u'j': u'じ',
        u't': u'たちつてとっ',
        u'd': u'だぢづでど',
        u'n': u'なにぬねのん',
        u'h': u'はひふへほ',
        u'p': u'ぱぴぷぺぽ',
        u'b': u'ばびぶべぼ',
        u'm': u'まみむめも',
        u'y': u'やゆよゃゅょ',
        u'r': u'らりるれろ',
        u'w': u'わゐうゑを',
        }
    
    # Social IME APIを呼び出して変換
    # 結果は配列で返す
    def convert(self, arg, predict):
        debug_method(_(u'convert %(arg)s, %(predict)s') % {'arg': arg, 'predict': predict})
        self.nRequest = self.nRequest + 1
        earg = arg.encode(self.SERVER_CHARSET)
        if predict:
            # 予測変換
            params = urllib.urlencode({'string': earg, 'charset': self.SERVER_CHARSET})
            request = self.SERVER+'api2/predict.php?%s' % params
            self.nPredict = self.nPredict + 1
            pass
        else:
            # 文節で変換
            params = urllib.urlencode({'string': earg, 'resize[0]': '+'+str(len(arg)), 'charset': self.SERVER_CHARSET})
            request = self.SERVER+'api/?%s' % params
            self.nDict = self.nDict + 1
            pass
        try:
            f = urllib.urlopen(request)
            result = f.read().strip()
            uresult = result.decode(self.SERVER_CHARSET)
            array = uresult.split(u'\t')
            debug_method(uresult)
            # 特殊文字対策
            newarray = []
            for x in array:
                if isContentAny(x, self.specialbefore):
                    r = u'(concat "%s")' % self.specialxlat(x)
                    pass
                else:
                    r = x
                    pass
                newarray.append(r)
                pass
            ret = newarray
            if ret:
                status = self.SERVER_FOUND
                pass
            else:
                status = self.SERVER_NOT_FOUND
                pass
        except IOError, e:
            debug_method(_('server error') + str(f.info()))
            ret = []
            status = self.SERVER_ERROR
            pass
        except:
            debug_method(_('unknown error') + str(f.info()))
            ret = []
            status = self.SERVER_ERROR
            pass
        return (status, ret)
    
    def do_0(self, arg):             # CLIENT_END
        debug_method(_('CLIENT_END:') + arg)
        self.close_connection = True
        return arg
    
    def do_1(self, arg):             # CLIENT_REQUEST
        '''
        通常のSKK変換要求を処理する
        「かな」のみまたは「英字」のみの場合は、要求された変換対象文字列を1文節として「かな漢字変換API」で変換する。
        「かな」+最後の文字が「英字」または「*」の場合に、要求された変換対象文字列全体を「予測変換API」で変換する。
        '''
        ## 予測変換フラグ
        predict = False
        ## 送りありフラグ
        okuri = ''
        # 送り仮名だったら予測変換
        # 全てASCII文字か日本語文字だったら文節変換
        if self.isalnum(arg):
            # 英単語っぽい？全て数字か英字のASCII文字
            predict = False
        else:
            if arg[-1] in u'abcdefgihjklmnopqrstuvwxyzABCDEFGIHJKLMNOPQRSTUVWXYZ*':
                # EBDIC backendでのワイルドカード変換と送り仮名ありのとき予測変換
                predict = True
                if arg[-1] == u'*':
                    # ワイルドカード文字は送り仮名ではない
                    pass
                else:
                    okuri = arg[-1]
                    pass
                arg = arg[:-1]
            else:
                # 送りなし
                predict = False
                pass
            pass
        (status, result) = self.convert(arg, predict)
        if status == self.SERVER_FOUND:
            if okuri:
                newresult = {}
                for r in result:
                    if r[-1] in self.okurimap[okuri]:
                        newresult[r[:-1]] = None
                        pass
                    pass
                if newresult:
                    ret = u'%s/%s/\n' %  (status, u'/'.join(newresult.keys()))
                    self.nOK = self.nOK + 1
                    pass
                else:
                    ret = u'%s/%s/\n' % (self.SERVER_NOT_FOUND, arg)
                    self.nFail = self.nFail + 1
                    pass
                pass
            else:
                ret = u'%s/%s/\n' %  (status, u'/'.join(result))
                self.nOK = self.nOK + 1
                pass
        else:
            ret = u'%s/%s/\n' % (status, arg)
            if status == self.SERVER_ERROR:
                self.nError = self.nError + 1
                pass
            else:
                self.nFail = self.nFail + 1
                pass
            pass
        return ret
    
    def do_2(self, arg):             # CLIENT_VERSION
        return self.VERSION
    
    def do_3(self, arg):             # CLIENT_HOST
        hostname = socket.gethostname()
        ipaddr = socket.gethostbyname(hostname)
        ret = u'%s:%s: ' % (hostname, ipaddr)
        return ret
    
    def do_4(self, arg):        # CLIENT_SERVER_COMPLETION
        lookup = False
        if arg[-1] == u'~':        # skk-lookup方式
            arg = arg[:-1]
            lookup = True
            pass
        # まずは予測変換させてみる
        (status, result) = self.convert(arg, True)
        if status == self.SERVER_FOUND:
            newresult = []
            if lookup:
                newresult = result
            else:
                for x in result:
                    if self.isHiragana(x):
                        newresult.append(x)
                        pass
                    pass
                pass
            if newresult:
                ret = u'%s %s \n' % (self.SERVER_FOUND, u' '.join(newresult))
                self.nOK = self.nOK + 1
            else:
                ret = u'%s %s \n' % (self.SERVER_NOT_FOUND, arg)
                self.nFail = self.nFail + 1
                pass
            pass
        else:
            ret = u'%s %s \n' % (status, arg)
            if status == self.SERVER_ERROR:
                self.nError = self.nError + 1
                pass
            else:
                self.nFail = self.nFail + 1
                pass
            pass
        return ret

    def do_UNKNOWN(self, arg):
        logging.warninng(_('unsupported command (%(arg)r)') % {'arg': arg})
        return self.SERVER_ERROR
    
    close_connection = False
    
    def handle_one_request(self):
        raw_request = self.connection.recv(self.COMBUFSIZE)
        if raw_request:
            try:
                urequest = raw_request.decode(self.CLIENT_CHARSET)
                mname = u'do_' + urequest[0]
                if not hasattr(self, mname):
                    self.do_UNKNOWN(urequest)
                    return
                method = getattr(self, mname)
                urequest = urequest.strip()
                uret = method(urequest[1:])
                ret = uret.encode(self.CLIENT_CHARSET)
            except:
                logging.exception(_('server error'))
                self.nError = self.nError + 1
                uret = self.SERVER_ERROR
                ret = uret.encode(self.CLIENT_CHARSET)
                pass
            debug_method(uret)
            self.connection.sendall(ret)
            pass
        else:
            self.close_connection = True
            pass
        return

    def handle(self):
        try:
            m = _('new connection %(client)s %(server)s\n') % {'client':self.client_address, 'server':self.server}
            logging.info(m)
            while not self.close_connection:
                try:
                    self.handle_one_request()
                except:
                    logging.exception(_('protocol handler error'))
                    self.close_connection = True
                    pass
                pass
            m = _('%(client)s request: %(req)d(%(dict)d/%(predict)d) error: %(error)d ok: %(ok)d fail: %(fail)d\n') % {'client':self.client_address, 'req':self.nRequest, 'dict':self.nDict, 'predict':self.nPredict, 'error':self.nError, 'ok':self.nOK, 'fail':self.nFail}
            logging.info(m)
            m = _('connection closed %(client)s %(server)s\n') % {'client':self.client_address, 'server':self.server}
            logging.info(m)
        except:
            logging.exception(_('PySocialSKKServ got exception'))
            pass
        return

    def setup(self):
        SocketServer.StreamRequestHandler.setup(self)
        # プロトコルハンドラループに突入する前に初期化すべき何かをここに記述
        self.specialxlat = buildTranslater(self.specialafter)
        pass

    def finish(self):
        SocketServer.StreamRequestHandler.finish(self)
        # その他にクリーンアップすべき何かがあれば、ここに記述
        pass

    pass                        # end of class

def daemonize(stdin='/dev/null', stdout='/dev/null', stderr='/dev/null', run='/dev/null'):
    '''
    デーモン化手続き
    Platform: UNIX
    '''
    # 1st fork
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)
            pass
        pass
    except OSError, (errno, errstr):
        logging.debug(_('1st fork failed: %s(%d)\n') % (errstr, errno))
        sys.exit(1)
        pass
 # separate from parent process
    os.chdir('/')
    os.umask(0)
    os.setsid()
        # 2nd fork
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)
            pass
        file(run, 'w').write(str(os.getpid()))
        pass
    except OSError, (errno, errstr):
        logging.debug(_('2nd fork failed: %s(%d)\n') % (errstr, errno))
        sys.exit(1)
        pass
        # daemonize complete
    for f in sys.stdout, sys.stderr:
        f.flush()
        pass
    si = open(stdin, 'r')
    so = open(stdout, 'a+', 0)
    se = open(stderr, 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())
    return

if sys.platform == 'win32':
    import win32serviceutil
    class PySocialSKKServService(win32serviceutil.ServiceFramework):
        _svc_name_ = 'PySocialSKKServ'
        _svc_display_name_ = 'Social IME SKKServ Service (Python)'
        _svc_description_ = 'SKKServ Social IME WebAPI binding'

        __service = None
        __serving = True
        def __init__(self, args):
            win32serviceutil.ServiceFramework.__init__(self, args)
            (lang, encoding) = locale.getdefaultlocale()
            print os.path.abspath('.')
            gt = gettext.translation('pysocialskkserv', '.', languages=[lang,])
            _ = gt.ugettext
            setupLogger()
            SocialSKKServer.daemon_threads = False
            self.__service = SocialSKKServer(('0.0.0.0', 5511), SocialSKKRequestHandler)
            pass

        # サービスに対するコマンドに対応するルーチン群
        def SvcStop(self):
            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
            pass

#         def SvcPause(self):
#             pass

#         def SvcContinue(self):
#             pass

#         def SvcInterrogate(self):
#             win32serviceutil.ServiceFramework.SvcInterrogate(self)
#             pass

        def SvcShutdown(self):
            self.__serving = False
            pass

        # サービスの開始が要求されると呼び出されるっぽい
        def SvcDoRun(self):
            while self.__serving:
                self.__service.handle_request()
            pass
    pass


if __name__ == '__main__':
    (lang, encoding) = locale.getdefaultlocale()
    # コンソールの実際のエンコーディングに設定する
    # PythonのバージョンによってはUnicodeErrorが出てしまう
    #sys.stdin = codecs.getreader(encoding)(sys.stdin)
    #sys.stdout = codecs.getwriter(encoding)(sys.stdout)
    #sys.stderr = codecs.getwriter(encoding)(sys.stderr)
    print lang, encoding
    import gettext
    gt = gettext.translation('pysocialskkserv', '.', languages=[lang,])
    _ = gt.ugettext
    parseoptions()
    setupLogger()
    if sys.platform == 'win32':
        # pywin32でサービス化できるらしい
        if len(sys.argv) > 1:
            try:
                err = win32serviceutil.HandleCommandLine(PySocialSKKServService)
                print err
                sys.exit(err)
            except NameError, e:
                logging.exception('NameError')
                logging.info(_('PySocialSKKServ service not installed. Run as termianl mode'))
                pass
            pass
        pass
    else:
        daemonize(run=options.run)
        pass
    logging.info(_('initializing PySocialSKKServ\n'))
    server = SocialSKKServer(('0.0.0.0', options.port), SocialSKKRequestHandler)
    logging.info(_('PySocialSKKServ service start\n'))
    server.serve_forever()
    pass
